/*
Copyright 2017-2023 Ian Jauslin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "constants.cpp"
#include "lyapunov.h"
#include <openblas64/cblas.h>
#include <openblas64/lapacke.h>
#include <math.h>
#include <stdlib.h>

#define MATSIZE (2*(K1*(2*K2+1)+K2))
#define USIZE (K1*(2*K2+1)+K2)

// compute Lyapunov exponents
int lyapunov(
  int K1,
  int K2,
  int N1,
  int N2,
  double final_time,
  double lyapunov_reset,
  double nu,
  double D_epsilon,
  double delta,
  double L,
  double adaptive_tolerance,
  double adaptive_factor,
  double max_delta,
  unsigned int adaptive_norm,
  _Complex double* u0,
  _Complex double* g,
  bool irreversible,
  bool keep_en_cst,
  double target_en,
  unsigned int algorithm,
  double starting_time,
  unsigned int nthreads
){
  double* lyapunov;
  double* lyapunov_avg;
  double* Du_prod;
  double* Du;
  _Complex double* u;
  _Complex double* prevu;
  _Complex double* tmp1;
  _Complex double* tmp2;
  _Complex double* tmp3;
  _Complex double* tmp4;
  _Complex double* tmp5;
  _Complex double* tmp6;
  _Complex double* tmp7;
  _Complex double* tmp8;
  _Complex double* tmp9;
  double* tmp10;
  double time;
  fft_vect fft1;
  fft_vect fft2;
  fft_vect ifft;

  // period
  // add 0.1 to ensure proper rounding
  uint64_t n=(uint64_t)((starting_time-fmod(starting_time, lyapunov_reset))/lyapunov_reset+0.1);

  lyapunov_init_tmps(&lyapunov, &lyapunov_avg, &Du_prod, &Du, &u, &prevu, &tmp1, &tmp2, &tmp3, &tmp4, &tmp5, &tmp6, &tmp7, &tmp8, &tmp9, &tmp10, &fft1, &fft2, &ifft, K1, K2, N1, N2, nthreads, algorithm);
  // copy initial condition
  copy_u(u, u0, K1, K2);

  // initialize Du_prod
  int i,j;
  for(i=0;i<MATSIZE;i++){
    for(j=0;j<MATSIZE;j++){
      Du_prod[i*MATSIZE+j]=0.;
    }
  }
  for(i=0;i<MATSIZE;i++){
    Du_prod[i*MATSIZE+i]=1.;
  }

  // store step (useful for adaptive step methods
  double prev_step=delta;
  double step=delta;
  double next_step=step;

  // init average
  for (i=0; i<MATSIZE; i++){
    lyapunov_avg[i]=0;
  }

  // save times at which Lyapunov exponents are computed
  double prevtime=starting_time;

  // iterate
  time=starting_time;
  while(final_time<0 || time<final_time){
    // copy before step
    copy_u(prevu, u, K1, K2);
    prev_step=step;

    ns_step(algorithm, u, K1, K2, N1, N2, nu, &step, &next_step, adaptive_tolerance, adaptive_factor, max_delta, adaptive_norm, L, g, time, starting_time, fft1, fft2, ifft, &tmp1, &tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, irreversible, keep_en_cst, target_en);

    // compute Jacobian
    // do not touch tmp1, it might be used in the next step
    ns_D_step(Du, prevu, u, K1, K2, N1, N2, nu, D_epsilon, prev_step, algorithm, adaptive_tolerance, adaptive_factor, max_delta, adaptive_norm, L, g, time, fft1, fft2, ifft, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9, irreversible, keep_en_cst, target_en);

    // multiply Jacobian
    // switch to column major order, so transpose Du
    cblas_dgemm(CblasColMajor, CblasNoTrans, CblasTrans, MATSIZE, MATSIZE, MATSIZE, 1., Du_prod, MATSIZE, Du, MATSIZE, 0., tmp10, MATSIZE);
    // copy back to Du_prod
    double* move=tmp10;
    tmp10=Du_prod;
    Du_prod=move;

    // increment time
    time+=step;

    // reset Jacobian
    if(time>(n+1)*lyapunov_reset){
      n++;

      // QR decomposition
      // do not touch tmp1, it might be used in the next step
      LAPACKE_dgerqf(LAPACK_COL_MAJOR, MATSIZE, MATSIZE, Du_prod, MATSIZE, tmp10);
      // extract eigenvalues (diagonal elements of Du_prod)
      for(i=0; i<MATSIZE; i++){
	lyapunov[i]=log(fabs(Du_prod[i*MATSIZE+i]))/(time-prevtime);
      }

      // sort lyapunov exponents
      qsort(lyapunov, MATSIZE, sizeof(double), compare_double);

      // average lyapunov
      for(i=0; i<MATSIZE; i++){
	// exclude inf
	if((! isinf(lyapunov[i])) && (time>starting_time)){
	  lyapunov_avg[i]=lyapunov_avg[i]*(prevtime-starting_time)/(time-starting_time)+lyapunov[i]*(time-prevtime)/(time-starting_time);
	}
      }

      // print
      for(i=0; i<MATSIZE; i++){
	printf("% .15e % .15e % .15e\n",time, lyapunov[i], lyapunov_avg[i]);
      }
      printf("\n");
      fprintf(stderr,"% .15e",time);
      // print largest and smallest lyapunov exponent
      if(MATSIZE>0){
	fprintf(stderr," % .15e % .15e\n", lyapunov[0], lyapunov[MATSIZE-1]);
      }

      // set Du_prod to Q:
      LAPACKE_dorgrq(LAPACK_COL_MAJOR, MATSIZE, MATSIZE, MATSIZE, Du_prod, MATSIZE, tmp10);

      // reset prevtime
      prevtime=time;
    }
  }

  lyapunov_free_tmps(lyapunov, lyapunov_avg, Du_prod, Du, u, prevu, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9, tmp10, fft1, fft2, ifft, algorithm);
  return(0);
}

// comparison function for qsort
int compare_double(const void* x, const void* y) {
  if (*(const double*)x<*(const double*)y) {
    return(-1);
  } else if (*(const double*)x>*(const double*)y) {
    return(+1);
  } else {
    return(0);
  }
}

// Jacobian of u_n -> u_{n+1}
int ns_D_step(
  double* out,
  _Complex double* un,
  _Complex double* un_next,
  int K1,
  int K2,
  int N1,
  int N2,
  double nu,
  double epsilon,
  double delta,
  unsigned int algorithm,
  double adaptive_tolerance,
  double adaptive_factor,
  double max_delta,
  unsigned int adaptive_norm,
  double L,
  _Complex double* g,
  double time,
  fft_vect fft1,
  fft_vect fft2,
  fft_vect ifft,
  _Complex double* tmp1,
  _Complex double* tmp2,
  _Complex double* tmp3,
  _Complex double* tmp4,
  _Complex double* tmp5,
  _Complex double* tmp6,
  _Complex double* tmp7,
  _Complex double* tmp8,
  bool irreversible,
  bool keep_en_cst,
  double target_en
){
  int lx,ly;
  int i;
  double step, next_step;

  for(lx=0;lx<=K1;lx++){
    for(ly=(lx>0 ? -K2 : 1);ly<=K2;ly++){
      // derivative in the real direction
      // finite difference vector
      for (i=0;i<USIZE;i++){
	if(i==klookup_sym(lx,ly,K2)){
	  tmp1[i]=un[i]+epsilon;
	}else{
	  tmp1[i]=un[i];
	}
      }
      // compute step 
      // reinitialize step
      step=delta;
      next_step=delta;
      ns_step(algorithm, tmp1, K1, K2, N1, N2, nu, &step, &next_step, adaptive_tolerance, adaptive_factor, max_delta, adaptive_norm, L, g, time, time, fft1, fft2, ifft, &tmp2, &tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, irreversible, keep_en_cst, target_en);
      // derivative
      for (i=0;i<USIZE;i++){
	// use row major order
	out[2*klookup_sym(lx,ly,K2)*MATSIZE+2*i]=(__real__ (tmp1[i]-un_next[i]))/epsilon;
	out[2*klookup_sym(lx,ly,K2)*MATSIZE+2*i+1]=(__imag__ (tmp1[i]-un_next[i]))/epsilon;
      }

      // derivative in the imaginary direction
      // finite difference vector
      for (i=0;i<USIZE;i++){
	if(i==klookup_sym(lx,ly,K2)){
	  tmp1[i]=un[i]+epsilon*I;
	}else{
	  tmp1[i]=un[i];
	}
      }
      // compute step 
      // reinitialize step
      step=delta;
      next_step=delta;
      ns_step(algorithm, tmp1, K1, K2, N1, N2, nu, &step, &next_step, adaptive_tolerance, adaptive_factor, max_delta, adaptive_norm, L, g, time, time, fft1, fft2, ifft, &tmp2, &tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, irreversible, keep_en_cst, target_en);
      // compute derivative
      for (i=0;i<USIZE;i++){
	// use row major order
	out[(2*klookup_sym(lx,ly,K2)+1)*MATSIZE+2*i]=(__real__ (tmp1[i]-un_next[i]))/epsilon;
	out[(2*klookup_sym(lx,ly,K2)+1)*MATSIZE+2*i+1]=(__imag__ (tmp1[i]-un_next[i]))/epsilon;
      }
    }
  }

  return(0);
}


int lyapunov_init_tmps(
  double ** lyapunov,
  double ** lyapunov_avg,
  double ** Du_prod,
  double ** Du,
  _Complex double ** u,
  _Complex double ** prevu,
  _Complex double ** tmp1,
  _Complex double ** tmp2,
  _Complex double ** tmp3,
  _Complex double ** tmp4,
  _Complex double ** tmp5,
  _Complex double ** tmp6,
  _Complex double ** tmp7,
  _Complex double ** tmp8,
  _Complex double ** tmp9,
  double ** tmp10,
  fft_vect* fft1,
  fft_vect* fft2,
  fft_vect* ifft,
  int K1,
  int K2,
  int N1,
  int N2,
  unsigned int nthreads,
  unsigned int algorithm
){
  ns_init_tmps(u, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9, fft1, fft2, ifft, K1, K2, N1, N2, nthreads, algorithm);

  *lyapunov=calloc(MATSIZE,sizeof(double));
  *lyapunov_avg=calloc(MATSIZE,sizeof(double));
  *Du_prod=calloc(MATSIZE*MATSIZE,sizeof(double));
  *Du=calloc(MATSIZE*MATSIZE,sizeof(double));
  *prevu=calloc(USIZE,sizeof(_Complex double));
  *tmp1=calloc(USIZE,sizeof(_Complex double));
  *tmp2=calloc(USIZE,sizeof(_Complex double));
  *tmp10=calloc(MATSIZE*MATSIZE,sizeof(double));

  return(0);
}

// release vectors
int lyapunov_free_tmps(
  double* lyapunov,
  double* lyapunov_avg,
  double* Du_prod,
  double* Du,
  _Complex double* u,
  _Complex double* prevu,
  _Complex double* tmp1,
  _Complex double* tmp2,
  _Complex double* tmp3,
  _Complex double* tmp4,
  _Complex double* tmp5,
  _Complex double* tmp6,
  _Complex double* tmp7,
  _Complex double* tmp8,
  _Complex double* tmp9,
  double* tmp10,
  fft_vect fft1,
  fft_vect fft2,
  fft_vect ifft,
  unsigned int algorithm
){
  free(tmp10);
  free(tmp2);
  free(tmp1);
  free(prevu);
  free(Du);
  free(Du_prod);
  free(lyapunov_avg);
  free(lyapunov);
  ns_free_tmps(u, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9, fft1, fft2, ifft, algorithm);

  return(0);
}
